import asyncio
import sys

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from typing import TYPE_CHECKING, Any, AsyncIterator
from ..messages.base import Message
from ..messages.commons import Audio, Text
from .base import Node
from .registry import NodeFactory

if TYPE_CHECKING:
    from google.cloud import texttospeech
    from google.api_core import exceptions
    from google.api_core.client_options import ClientOptions

try:
    from google.cloud import texttospeech
    from google.api_core import exceptions
    from google.api_core.client_options import ClientOptions

    GOOGLE_CLOUD_TEXTTOSPEECH_AVAILABLE = True
except ImportError:
    GOOGLE_CLOUD_TEXTTOSPEECH_AVAILABLE = False


@NodeFactory.register("tts")
class TTSNode(Node[Text, Audio]):
    def __init__(
        self,
        input_channel: str,
        output_channel: str,
        api_key: str,
        rate: int,
        redis_url: str,
    ) -> None:
        if not GOOGLE_CLOUD_TEXTTOSPEECH_AVAILABLE:
            raise ImportError(
                "Google Cloud Text-to-Speech is not available. Please install"
                "aact with `pip install aact[google]` to use the TTSNode."
            )
        super().__init__(
            input_channel_types=[(input_channel, Text)],
            output_channel_types=[(output_channel, Audio)],
            redis_url=redis_url,
        )
        self.input_channel = input_channel
        self.output_channel = output_channel
        self.client = texttospeech.TextToSpeechAsyncClient(
            client_options=ClientOptions(api_key=api_key)  # type: ignore[no-untyped-call]
        )
        self.voice = texttospeech.VoiceSelectionParams(
            language_code="en-US", ssml_gender=texttospeech.SsmlVoiceGender.NEUTRAL
        )
        self.audio_config = texttospeech.AudioConfig(
            audio_encoding=texttospeech.AudioEncoding.LINEAR16, sample_rate_hertz=rate
        )
        self.queue = asyncio.Queue[str]()
        self.shutdown = asyncio.Event()

    async def synthesize(self) -> None:
        while not self.shutdown.is_set():
            text = await self.queue.get()
            try:
                input_text = texttospeech.SynthesisInput(text=text)
                response = await self.client.synthesize_speech(
                    request={
                        "input": input_text,
                        "voice": self.voice,
                        "audio_config": self.audio_config,
                    }
                )
                await self.r.publish(
                    self.output_channel,
                    Message[Audio](
                        data=Audio(audio=response.audio_content)
                    ).model_dump_json(),
                )
            except exceptions.GoogleAPICallError as e:
                print(f"Error during speech synthesis: {e}")

    async def __aenter__(self) -> Self:
        self.task = asyncio.create_task(self.synthesize())
        return await super().__aenter__()

    async def __aexit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        self.shutdown.set()
        return await super().__aexit__(exc_type, exc_value, traceback)

    async def event_handler(
        self, input_channel: str, text: Message[Text]
    ) -> AsyncIterator[tuple[str, Message[Audio]]]:
        if input_channel != self.input_channel:
            raise ValueError(f"Unexpected input channel: {input_channel}")
            yield "", Message[Audio](data=Audio(audio=b""))
        else:
            await self.queue.put(text.data.text)



from typing import AsyncIterator
from aact import Node, NodeFactory
from aact.messages import Message, Text

from openai import OpenAI


@NodeFactory.register("gpt4_text_chatbot_node")
class GPT4TextChatbotNode(Node[Text, Text]):
    def __init__(self, input_channel: str, output_channel: str, redis_url: str):
        super().__init__(
            input_channel_types=[(input_channel, Text)],
            output_channel_types=[(output_channel, Text)],
            redis_url=redis_url,
        )
        self.input_channel = input_channel
        self.output_channel = output_channel
        self.client = OpenAI()

    async def event_handler(
        self, channel: str, message: Message[Text]
    ) -> AsyncIterator[tuple[str, Message[Text]]]:
        match channel:
            case self.input_channel:
                response = self.client.chat.completions.create(
                    model="gpt-4o-mini",
                    messages=[
                        {"role": "system", "content": "You are a helpful assistant."},
                        {"role": "user", "content": message.data.text},
                    ],
                )
                yield (
                    self.output_channel,
                    Message[Text](data=Text(text=response.choices[0].message.content)),
                )
            case _:
                raise ValueError(f"Unexpected channel: {channel}")



#!/usr/bin/env python3
from datetime import datetime
from typing import List, Tuple
from uuid import uuid4

from nicegui import ui

messages: List[Tuple[str, str, str, str]] = []


@ui.refreshable
def chat_messages(own_id: str) -> None:
    if messages:
        for user_id, avatar, text, stamp in messages:
            ui.chat_message(
                text=text, stamp=stamp, avatar=avatar, sent=own_id == user_id
            )
    else:
        ui.label("No messages yet").classes("mx-auto my-36")
    ui.run_javascript("window.scrollTo(0, document.body.scrollHeight)")


@ui.page("/")
async def main():
    def send() -> None:
        stamp = datetime.utcnow().strftime("%X")
        messages.append((user_id, avatar, text.value, stamp))
        text.value = ""
        chat_messages.refresh()

    user_id = str(uuid4())
    avatar = f"https://robohash.org/{user_id}?bgset=bg2"

    ui.add_css(
        r"a:link, a:visited {color: inherit !important; text-decoration: none; font-weight: 500}"
    )
    with ui.footer().classes("bg-white"), ui.column().classes(
        "w-full max-w-3xl mx-auto my-6"
    ):
        with ui.row().classes("w-full no-wrap items-center"):
            with ui.avatar().on("click", lambda: ui.navigate.to(main)):
                ui.image(avatar)
            text = (
                ui.input(placeholder="message")
                .on("keydown.enter", send)
                .props("rounded outlined input-class=mx-3")
                .classes("flex-grow")
            )
        ui.markdown("simple chat app built with [NiceGUI](https://nicegui.io)").classes(
            "text-xs self-end mr-8 m-[-1em] text-primary"
        )

    await (
        ui.context.client.connected()
    )  # chat_messages(...) uses run_javascript which is only possible after connecting
    with ui.column().classes("w-full max-w-2xl mx-auto items-stretch"):
        chat_messages(user_id)


if __name__ in {"__main__", "__mp_main__"}:
    ui.run(native=True)
